#!/usr/bin/env python
from optparse import OptionParser
from graphviz import Digraph
from kazoo.client import KazooClient
from concord_cli.generated.concord.internal.thrift.ttypes import *

import json
import logging

from thrift.protocol import TJSONProtocol, TBinaryProtocol
from thrift.transport import TTransport

from concord_cli.utils import *

logging.basicConfig()
logger = logging.getLogger('cmd.tracer')
logger.setLevel(logging.INFO)

def generate_options():
    usage = "usage: %prog [options] arg"
    parser = OptionParser(usage)
    parser.add_option("--zookeeper", dest="zookeeper", default="localhost:2181",
                      action="store", help="i.e: 1.2.3.4:2181,2.3.4.5:2181")
    parser.add_option("--file", help="output file",
                      default="trace_graph", action="store", dest="filename")
    parser.add_option("--scheduler-path", help="zookeeper path, i.e.: /concord",
                      default="/concord", action="store", dest="zk_path")
    parser.add_option("--trace-id", help="any trace id, i.e.: 1121345",
                      action="store", dest="trace_id")
    parser.add_option("--scheduler-address", help="i.e.: 1.2.3.4:11219",
                      action="store", dest="scheduler")
    return parser

def validate_options(options, parser):
    default_options(options)
    if not options.trace_id:
        parser.error("need to specify trace id")
    if not options.scheduler:
        parser.error("need to specify scheduler_address")

def gen_trace(options):
    logger.info("what is my ip? %s" % options.scheduler)
    (addr, port) = options.scheduler.split(":")
    trace_cli = get_trace_service_client(addr, 11219)
    logger.info("about to get trace id: %s" % options.trace_id)
    logger.debug(dir(trace_cli))
    spans = trace_cli.getTrace(long(options.trace_id))
    dot = Digraph(comment='Concord Systems',
                  format='svg',
                  engine='dot',
                  graph_attr={
                      "overlap":"false"
                  },
                  edge_attr={
                      "overlap":"false",
                      "splines":"true",
                  },
                  node_attr={"shape":"rectangle",
                             "overlap":"false",
                             "fontname":"Arial",
                             "fontsize":"16"})

    times = {}
    for span in spans:
        for ann in span.annotations:
            key = str(span.id) + "_" + str(ann.type)
            times[key] = ann.timestamp

    for span in spans:
        for ann in span.annotations:
            ant = str(
                AnnotationType. _VALUES_TO_NAMES[ann.type] +
                "\n" + ann.host.ip + ":" + str(ann.host.port)
            )
            parent = str(span.id)
            child = str(span.parentId)
            dot.node(parent, span.name)
            dot.node(child)
            if ann.type == AnnotationType.SERVER_SEND:
                key = str(span.id) + "_" + str(AnnotationType.SERVER_RECV)
                if times.has_key(key):
                    t = times[key]
                    ant += ", " + str(t - ann.timestamp) + "ms"
                dot.edge(parent, child, label=ant)
            elif ann.type == AnnotationType.SERVER_RECV:
                key = str(span.id) + "_" + str(AnnotationType.SERVER_SEND)
                if times.has_key(key):
                    t = times[key]
                    ant += ", " + str(ann.timestamp - t) + "ms"
                dot.edge(child, parent, label=ant)
            elif ann.type == AnnotationType.CLIENT_RECV:
                key = str(span.id) + "_" + str(AnnotationType.CLIENT_SEND)
                if times.has_key(key):
                    t = times[key]
                    ant += ", " + str(ann.timestamp - t) + "ms"
                dot.edge(parent, child, label=ant)
            elif ann.type == AnnotationType.CLIENT_SEND:
                key = str(span.id) + "_" + str(AnnotationType.CLIENT_RECV)
                if times.has_key(key):
                    t = times[key]
                    ant += ", " + str(t - ann.timestamp) + "ms"
                dot.edge(child, parent, label=ant)
            else:
                dot.edge(parent,child,label='unknown')

    logger.info("Graph generated, rendering now")
    dot.render(options.filename, view=True, cleanup=True)


def main():
    parser = generate_options()
    (options, args) = parser.parse_args()
    validate_options(options,parser)
    gen_trace(options)

if __name__ == "__main__":
    main()



import os
import sys
import json
import argparse
from functools import partial
from concord_cli.utils import *

CONCORD_DEFAULTS = { 'zookeeper_path' : '/concord',
                     'zookeeper_hosts' : 'localhost:2181',
                     'scheduler_address' : 'localhost:11219' }

def generate_options():
    parser = argparse.ArgumentParser()
    subparser = parser.add_subparsers()

    # Init command takes two optional arguments 'location' and 'parameters'
    init_parser = subparser.add_parser('init', help='Create concord cli defaults')
    init_parser.add_argument('-c', '--config', default="~",
                             help="location of settings file, defaults to ~")
    init_parser.add_argument('-p', '--parameters', nargs='+', type=str,
                             help='Custom cli defaults, i.e. k1=v1 k2=v2')
    init_parser.set_defaults(which='init')

    # Set command takes one positional argument 'parameters'
    set_parser = subparser.add_parser('set', help='Add/Modify concord cli defaults')
    set_parser.add_argument('parameters', nargs='+', type=str,
                            help='Custom cli defaults, i.e. k1=v1 k2=v2')
    set_parser.set_defaults(which='set')

    show_parser = subparser.add_parser('show', help='Show concord cli defaults')
    show_parser.set_defaults(which='show')

    return parser

def fetch_config(location, callback, default_path=None):
    """ Attempts to find file starting at location. Upon success the file
    will be opened and contents passed into callback"""
    filepath = find_config(location, CONCORD_FILENAME)
    if filepath is None:
        print 'While searching recursively upwards (starting from %s), ' \
            'a configuration file could not be found...' % location
        return default_path

    with open(filepath, 'r') as datafile:
        config_data = json.load(datafile)

    return callback(config_data, filepath)

def write_defaults(new_defaults, current, writepath):
    """ Writes 'new_defaults' to disk at 'writepath' using 'current' for
    default valies that 'new_defaults' may be omitting"""
    def validate_config_params(config_params):
        valid_config_keys = CONCORD_DEFAULTS.keys()
        def validate(key):
            return key in valid_config_keys
        return all(map(validate, config_params.keys()))

    if new_defaults is not None:
        if not validate_config_params(new_defaults):
            raise Exception("Attempting to set unrecognized parameter")
        current.update(new_defaults)

    with open(writepath , 'w') as outfile:
        json.dump(current, outfile)

def init(location, new_defaults=None):
    """ Creates config_file in location, asking user to overwrite if already
    exists. Then prompts user for default values and writes to disk if -p
    option has not been used"""
    def duplicate_prompt(config_data, fpath):
        print 'config_file already exits in', fpath
        inpt = raw_input('Overwrite? [y|n] ')
        return fpath if inpt == "y" else sys.exit(1)

    default_path = os.path.join(os.path.expanduser(location), CONCORD_FILENAME)
    filepath = fetch_config(os.getcwd(), duplicate_prompt, default_path)

    if new_defaults is None:
        def prompt_default(k, v):
            inpt = raw_input('-- for key "%s" value will be ["%s"]: ' % (k,v))
            return v if inpt == "" else inpt

        print '\nCreating %s, select or overwrite default values: ' % filepath
        print 'When prompted, hit enter to take the default or enter a new value'
        new_defaults = { k:prompt_default(k,v) for k,v in CONCORD_DEFAULTS.iteritems() }

    write_defaults(new_defaults, CONCORD_DEFAULTS.copy(), filepath)
    print '\nWrite successful, these are the new settings: '
    fetch_config(filepath, show)

def show(config_data, filepath):
    """ Prints the contents of the config_file to stdout"""
    def pretty_print(kvpair):
        print '%s=%s' % kvpair
    print 'Concord configuration data: ', filepath
    map(pretty_print, config_data.iteritems())

def main():
    parser = generate_options()
    options = parser.parse_args()
    if options.which == 'show':
        fetch_config(os.getcwd(), show)
    else:
        argsdict = pairs_todict(options.parameters)
        if options.which == 'init':
            init(options.config, argsdict)
        elif options.which == 'set':
            fetch_config(os.getcwd(), partial(write_defaults, argsdict))

if __name__ == '__main__':
    main()



import os
import json
import logging
from thrift import Thrift
from kazoo.client import KazooClient
from concord_cli.generated.concord.internal.thrift.ttypes import *
from concord_cli.generated.concord.internal.thrift import (
    BoltTraceAggregatorService,
    BoltSchedulerService
)
from thrift.protocol import (
    TJSONProtocol, TBinaryProtocol
)
from thrift.transport import (
    TSocket,TTransport
)

logging.basicConfig()
logger = logging.getLogger('cmd.utils')
logger.setLevel(logging.INFO)

CONCORD_FILENAME = '.concord.cfg'

class ContextDirMgr:
    def __init__(self, path):
        self.new_dir = os.path.dirname(os.path.abspath(path))

    def __enter__(self):
        self.old_dir = os.getcwd()
        os.chdir(self.new_dir)

    def __exit__(self, value, type, traceback):
        os.chdir(self.old_dir)

def bytes_to_thrift(bytes, thrift_struct):
    transportIn = TTransport.TMemoryBuffer(bytes)
    protocolIn = TBinaryProtocol.TBinaryProtocol(transportIn)
    thrift_struct.read(protocolIn)

def thrift_to_json(thrift_struct):
    return json.dumps(thrift_struct, default=lambda o: o.__dict__,
            sort_keys=True, indent=4)

def get_zookeeper_master_ip(zkurl, zkpath):
    logger.info("Connecting to:%s" % zkurl)
    zk = KazooClient(hosts=zkurl)
    ip = ""
    try:
        logger.debug("Starting zk connection")
        zk.start()
        if not zk.exists(zkpath):
            logger.error('Path on zk doesn\'t exist')
            return ip
        logger.debug("Serializing TopologyMetadata() from %s" % zkpath)
        data, stat = zk.get(zkpath + "/masterip")
        logger.debug("Status of 'getting' %s/masterip: %s" % (zkpath, str(stat)))
        ip = str(data)
    except Exception as e:
        logger.exception(e)
    finally:
        logger.debug("Closing zk connection")
        zk.stop()
    return ip

def get_zookeeper_metadata(zkurl, zkpath):
    logger.info("Connecting to: %s" % zkurl)
    zk = KazooClient(hosts=zkurl)
    meta = TopologyMetadata()
    try:
        logger.debug("Starting zk connection")
        zk.start()
        if not zk.exists(zkpath):
            logger.error('Path on zk doesn\'t exist')
            return None
        logger.debug("Serializing TopologyMetadata() from %s" % zkpath)
        data, stat = zk.get(zkpath)
        logger.debug("Status of 'getting' %s: %s" % (zkpath, str(stat)))
        bytes_to_thrift(data, meta)
    except Exception as e:
        logger.exception(e)
        return None
    finally:
        logger.debug("Closing zk connection")
        zk.stop()

    return meta

def tproto(ip, port):
    socket = TSocket.TSocket(ip, port)
    transport = TTransport.TFramedTransport(socket)
    protocol = TBinaryProtocol.TBinaryProtocol(transport)
    return (protocol, transport)

def get_sched_service_client(ip, port):
    (protocol, transport) = tproto(ip, port)
    client = BoltSchedulerService.Client(protocol)
    transport.open()
    return client

def get_trace_service_client(ip, port):
    (protocol, transport) = tproto(ip, port)
    client = BoltTraceAggregatorService.Client(protocol)
    transport.open()
    return client

def flatten(xs):
    return reduce(lambda m, x: m + x, xs,[])

def pairs_todict(kvpair_list):
    """ Transform list -> dict i.e. [a=b, c=d] -> {'a':'b', 'c':'d'}"""
    return None if kvpair_list is None else \
    { k:v for k, v in map(lambda c: c.split('='), kvpair_list) }

def find_config(src, config_file):
    """ recursively searches .. until it finds a file named config_file
    will return None in the case of no matches or the abspath if found"""
    filepath = os.path.join(src, config_file)
    if os.path.isfile(filepath):
        return filepath
    elif src == '/':
        return None
    else:
        return find_config(os.path.dirname(src), config_file)

def default_options(opts):
    location = find_config(os.getcwd(), CONCORD_FILENAME)
    if location is None:
        return
    with open(location, 'r') as data_file:
        config_data = json.load(data_file)
    opts_methods = dir(opts)
    if 'zookeeper' in opts_methods:
        opts.zookeeper = config_data['zookeeper_hosts']
    if 'zk_path' in opts_methods:
        opts.zk_path = config_data['zookeeper_path']
    if 'scheduler' in opts_methods:
        opts.scheduler = config_data['scheduler_address']



import os
import unittest
from concord_cli.generated.concord.internal.thrift.ttypes import *
from concord_cli.print_graph import *
from testing_utilities import *

class TestPrintGraph(unittest.TestCase):

    def setUp(self):
        # Retrieve static topology metadata
        with open(test_filepath('test_topology_metadata')) as data_file:
            self.test_topology_metadata = data_file.read()

        # Retrieve result pdf generated by static topology metadata
        with open(test_filepath('test_topology_visual_graph')) as data_file:
            self.test_topology_graph = data_file.read()

        self.stubbed_metadata = TopologyMetadata()
        bytes_to_thrift(self.test_topology_metadata, self.stubbed_metadata)

    def test_print_dot(self):
        # Give print dot a filepath to store its result into
        temp_file_name = test_filepath("temporary")
        print_dot(self.stubbed_metadata, temp_file_name)

        # Read the contents of the data generated by print_dot into memory
        temp_file_name = temp_file_name + '.pdf'
        with open(temp_file_name) as data_file:
            test_data = data_file.read()

        try:
            # Assert that the file written by print_dot matches with test pdf
            self.assertEquals(test_data, self.test_topology_graph)
        finally:
            os.unlink(temp_file_name)



#!/usr/bin/env python
import sys
import thrift
import json
import logging
import argparse
from concord_cli.utils import *
from terminaltables import AsciiTable

logging.basicConfig()
logger = logging.getLogger('cmd.kill_task')
logger.setLevel(logging.INFO)

def generate_options():
    parser = argparse.ArgumentParser()
    parser.add_argument("-p"
                        ,"--zk_path"
                        ,metavar="zookeeper-path"
                        ,help="Path of concord topology on zk"
                        ,default="/concord"
                        ,action="store")
    parser.add_argument("-z"
                        ,"--zookeeper"
                        ,metavar="zookeeper-hosts"
                        ,help="i.e: 1.2.3.4:2181,2.3.4.5:2181"
                        ,default="localhost:2181"
                        ,action="store")
    parser.add_argument("-t"
                        ,"--task_id"
                        ,help="The id of task to kill"
                        ,action="store")
    parser.add_argument("-a"
                        ,"--all"
                        ,help="Kill all computations"
                        ,action="store_true")
    return parser

def validate_options(options, parser):
    if options.all and options.task_id:
        parser.error('You are using task_id and passing the all flag')
    default_options(options)

def kill(zookeeper, zk_path, task_ids):
    if len(task_ids) == 0:
        logger.info('Kill received an empty list of task_ids')
        return
    logger.info("Getting master ip from zookeeper")
    ip = get_zookeeper_master_ip(zookeeper, zk_path)
    logger.info("Found leader at: %s" % ip)
    (addr, port) = ip.split(":")
    logger.info("Initiating connection to scheduler")
    cli = get_sched_service_client(addr,int(port))
    logger.info("Sending request to scheduler")
    try:
        for task_id in task_ids:
            logger.info('Sending request to kill task: %s', task_id)
            cli.killTask(task_id)
    except thrift.Thrift.TApplicationException as e:
        leftover = task_ids[task_ids.index(task_id):]
        logger.error("Error killing task: %s" % task_id)
        logger.error("Tasks that could not be killed... %s" % ", ".join(leftover))
        logger.exception(e)
    logger.info("Done sending request to server")

def collect_taskids(zookeeper, zkpath):
    """ Querys zk for topology metadata and returns a list of task_ids"""
    meta = get_zookeeper_metadata(zookeeper, zkpath)
    if meta is None or len(meta.computations) == 0:
        return []

    # Obtain taskids from list of nodes from list of computations then flatten
    return flatten([list(d.taskId for d in c.nodes) \
                    for c in meta.computations.values()])

def parse_input(iterable, user_input):
    """ Parses user input and either will abort or return the data at
    a given index or indicies or range"""
    def token_expand(token):
        if token.isdigit():
            return [int(token)]
        # Otherwise, we are dealing with a range
        inclusive_range = token.split('..')
        if len(inclusive_range) == 2:
            return range(int(inclusive_range[0]), int(inclusive_range[1]) + 1)
        raise Exception('Unexpected token detected')

    tokens = user_input.split(',')
    if len(tokens) == 1:
        if tokens[0] == 'quit' or tokens[0] == 'q':
            sys.exit(1)
        elif tokens[0] == 'all' or tokens[0] == 'a':
            return iterable

    # Expand ranges, flatten results, then insert into set to remove dups
    indicies = set(flatten(map(token_expand, tokens)))

    # Returns desired choices performing bounds check at the same time
    def fetch_index(index):
        if index < 1 or index > len(iterable):
            raise Exception('Index out of bounds')
        return iterable[index-1]
    return map(fetch_index, indicies)

def prompt_selection(iterable, heading, printer):
    """ Takes 'iterable' and displays it in an ascii table using printable
    method to determine how to transform the data. Will also prompt user for
    a selection and return the blob to the caller"""
    def preprinter(value_pair):
        idx, data = value_pair
        return [str(idx)] + printer(data)

    # Create table
    heading.insert(0, 'Idx')
    table_data = zip(xrange(1, len(iterable) + 1), iterable)
    table_data = map(preprinter, table_data)
    table_data.insert(0, heading)

    # Print table in nice format, wait for user selection
    print AsciiTable(table_data).table
    inpt = ""
    while inpt == "":
        inpt = raw_input('Selection: [1..%d, 1,x,y,..., quit(q), all(a)]:'\
                         ' ' % len(iterable))

    return parse_input(iterable, inpt)

def prompt_nodes(pcomp_metadata):
    def pcomp_printable(pcomp_metadata):
        return [pcomp_metadata.taskId,
                str(pcomp_metadata.cpus),
                str(pcomp_metadata.mem)]

    print '\nSelect a node to terminate:'
    nodedetail_header = ['Task id', 'cpus', 'mem']
    selection = prompt_selection(pcomp_metadata, nodedetail_header, pcomp_printable)
    return map(lambda c: c.taskId, selection)

def prompt_computations(pcomp_layout):
    def layout_printable(pcomp_layout):
        def pretty_md(stream_metadata):
            return "\n".join(map(lambda c: '(%s <-> %s)' % (c.name, c.grouping),
                                 stream_metadata))
        return [pcomp_layout.name,
                pretty_md(pcomp_layout.istreams),
                "\n".join(pcomp_layout.ostreams)]

    print '\nSelect a computation to inspect:'
    header = ['Computation name', 'istreams/grouping', 'ostreams/grouping']
    selection = prompt_selection(pcomp_layout, header, layout_printable)
    return flatten(map(lambda c: c.nodes, selection))

def interactive_mode(zookeeper, zk_path):
    """ presents to the user an easy to use prompt so that they may selectively
    search through computations and eventually choose a concord node to kill"""
    print 'Querying zookeeper for cluster topology...'
    meta = get_zookeeper_metadata(zookeeper, zk_path)

    # Exit on failure, or if no computations exist
    if meta == None:
        logger.critical('Could not connect to zk')
        return
    elif meta.computations == None or len(meta.computations) == 0:
        logger.info('No computations to report')
        return

    selection = meta.computations.values()
    user_path = [prompt_computations, prompt_nodes]
    for current_action in user_path:
        selection = current_action(selection)

    # Terminate selected node
    kill(zookeeper, zk_path, selection)

def main():
    parser = generate_options()
    options = parser.parse_args()
    validate_options(options, parser)

    if not options.task_id and not options.all:
        interactive_mode(options.zookeeper, options.zk_path)
    elif not options.all:
        kill(options.zookeeper, options.zk_path, options.task_id)
    elif not options.task_id:
        kill(options.zookeeper, options.zk_path,
             collect_taskids(options.zookeeper, options.zk_path)
)


if __name__ == "__main__":
    main()



import unittest
import json
import tempfile
import shutil
import re
from concord_cli.deploy import *
from testing_utilities import *
from subprocess import call

class StubParser:
    def error(self, error_string):
        raise RuntimeError(error_string)


class TestDeployScript(unittest.TestCase):

    def setUp(self):
        self.requiredKeys = ['compress_files', 'executable_name', 'computation_name',
                             'zookeeper_hosts', 'zookeeper_path']
        self.validKeys = ["executable_arguments", "docker_container",
                          "fetch_url", "mem", "disk", "cpus",
                          "framework_v_module", "instances",
                          "framework_logging_level", "environment_variables",
                          "zookeeper_hosts", "exclude_compress_files",
                          "update_binary", "execute_as_user",
                          "docker_container"]
        self.filename = 'test_deploy_file'
        self.improperFormatFilename = 'test_deploy_file_incorrect'
        self.stub = StubParser()

    def test_validate_json_raw_config(self):
        valid_dict = dict.fromkeys(self.validKeys)
        req_dict = dict.fromkeys(self.requiredKeys)

        # Test that an empty config file fails
        self.assertRaises(RuntimeError, validate_json_raw_config, {}, self.stub)

        # Even if all optional keys exist, method should fail
        self.assertRaises(RuntimeError,
                          validate_json_raw_config,
                          valid_dict,
                          self.stub)

        # Even a dictionary with at least one invalid key fails
        test_config_dict = req_dict.copy()
        test_config_dict['Invalid Key'] = 'Invalid Value'
        self.assertRaises(RuntimeError,
                          validate_json_raw_config,
                          test_config_dict,
                          self.stub)

        # A dictionary must have all of the required keys
        test_config_dict = req_dict.copy()
        test_config_dict.pop("compress_files")
        self.assertRaises(RuntimeError,
                          validate_json_raw_config,
                          test_config_dict,
                          self.stub)

        # A dictionary with only the required keys passes
        validate_json_raw_config(req_dict, self.stub)


    def test_parse_file(self):
        # Assert that parser fails if file isn't in JSON format
        self.assertRaises(ValueError,
                          parseFile,
                          test_filepath(self.improperFormatFilename),
                          self.stub)

        # Read test configuration file
        data = {}
        abs_path = test_filepath(self.filename)
        with open(abs_path) as data_file:
            data = json.load(data_file)

        cData = DEFAULTS.copy()
        cData.update(data)

        # Retrieve configuration options from method
        conf = parseFile(abs_path, self.stub)

        # Assert that the executable name exists as one of the compress_files
        self.assertTrue(conf["executable_name"] in conf["compress_files"])

        # Assert that the options dictionary contains expected key/value pairs
        self.assertDictEqual(cData, conf)

    def test_tar_file_list(self):
        total_files = [ 'test_directory/a1', 'test_directory/a2',
                        'test_directory/a3',  'test_directory/a4',
                        'test_directory/a5',  'test_directory/a6',
                        'test_directory/b1', 'test_directory/b2',
                        'test_directory/subdir/a1', 'test_directory/subdir/b1']

        # Assert that nothing gets inserted (using empty whitelist)
        white_list = []
        black_list = [ '.*a6' ]
        succ_list = []
        self.tar_tester_driver(total_files, white_list, black_list, succ_list)

        # Assert that nothing gets inserted (using blacklist with greedy glob)
        white_list = list(total_files)
        black_list = [ '.*' ]
        succ_list = []
        self.tar_tester_driver(total_files, white_list, black_list, succ_list)

        # Assert that pattern matching is working to remove correct file kinds
        # and also verify that matching works even in subdirectories
        white_list = list(total_files)
        black_list = [ '.*a' ]
        succ_list = [ 'test_directory/b1', 'test_directory/b2',
                      'test_directory/subdir/b1']
        self.tar_tester_driver(total_files, white_list, black_list, succ_list)


    def tar_tester_driver(self, total_files, white_list, black_list, successful):
        # Create actual temporary directory
        temp_dirname = tempfile.mkdtemp();

        # Pre-pend all strings in both lists with 'temp_dirname/'
        def prepend_dirname(item):
            return temp_dirname + '/' + item

        white_list = map(prepend_dirname, white_list)
        total_files = map(prepend_dirname, total_files)
        successful = map(prepend_dirname, successful)

        try:
            # Create new files from fake data
            for f in total_files:
                create_temporary_file(f)

            # Assert that only elements on the white list remain
            all_include = tar_file_list(white_list, black_list)
            all_include.sort()
            successful.sort()
            self.assertListEqual(all_include, successful)
        finally:
            # Remove temporary directory and all of its related content
            shutil.rmtree(temp_dirname)
